import json
import logging
import logging.config
import os
import pathlib
import uuid
from urllib.request import urlopen

import coloredlogs
import yaml
from common_utils_py.ddo.ddo import DDO
from common_utils_py.http_requests.requests_session import get_requests_session
from contracts_lib_py.utils import get_account

from nevermined_sdk_py.gateway.gateway_provider import GatewayProvider
from nevermined_sdk_py.gateway.gateway import Gateway
from nevermined_sdk_py.secret_store.secret_store import SecretStore
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from nevermined_sdk_py.nevermined.nevermined import Nevermined
from nevermined_sdk_py.secret_store.secret_store_provider import SecretStoreProvider
from tests.resources.mocks.gateway_mock import GatewayMock
from tests.resources.mocks.secret_store_mock import SecretStoreMock

PUBLISHER_INDEX = 1
CONSUMER_INDEX = 0


def get_resource_path(dir_name, file_name):
    base = os.path.realpath(__file__).split(os.path.sep)[1:-1]
    if dir_name:
        return pathlib.Path(os.path.join(os.path.sep, *base, dir_name, file_name))
    else:
        return pathlib.Path(os.path.join(os.path.sep, *base, file_name))


def init_ocn_tokens(nevermined, account, amount=100):
    nevermined.accounts.request_tokens(account, amount)
    Keeper.get_instance().token.token_approve(
        Keeper.get_instance().dispenser.address,
        amount,
        account
    )


def get_publisher_account():
    return get_account(0)


def get_consumer_account():
    return get_account(1)


def get_publisher_instance(init_tokens=True, use_ss_mock=True, use_gateway_mock=True):
    nevermined = Nevermined()
    account = get_publisher_account()
    nevermined.main_account = account
    if init_tokens:
        init_ocn_tokens(nevermined, nevermined.main_account)
    if use_ss_mock:
        SecretStoreProvider.set_secret_store_class(SecretStoreMock)
    else:
        SecretStoreProvider.set_secret_store_class(SecretStore)
    if use_gateway_mock:
        GatewayProvider.set_gateway_class(GatewayMock)
    else:
        Gateway.set_http_client(get_requests_session())
        GatewayProvider.set_gateway_class(Gateway)

    return nevermined


def get_consumer_instance(init_tokens=True, use_ss_mock=True, use_gateway_mock=True):
    nevermined = Nevermined()
    account = get_consumer_account()
    nevermined.main_account = account
    if init_tokens:
        init_ocn_tokens(nevermined, nevermined.main_account)
    if use_ss_mock:
        SecretStoreProvider.set_secret_store_class(SecretStoreMock)
    else:
        SecretStoreProvider.set_secret_store_class(SecretStore)
    if use_gateway_mock:
        GatewayProvider.set_gateway_class(GatewayMock)
    else:
        Gateway.set_http_client(get_requests_session())
        GatewayProvider.set_gateway_class(Gateway)

    return nevermined


def _get_asset(url):
    return DDO(json_text=get_assset_json_text(url))


def get_assset_json_text(url):
    return json.dumps(json.loads(urlopen(url).read().decode('utf-8')))


def get_ddo_sample():
    return _get_asset(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/metadata/v0.1/ddo-example-access.json")


def get_algorithm_ddo():
    return json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/1a853613bd9e9d633474e571d3225378a37b6451/docs/architecture/specs/examples/metadata/v0.1/ddo-example-algorithm.json").read().decode(
        'utf-8'))


def get_workflow_ddo():
    return json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/metadata/v0.1/ddo-example-workflow.json").read().decode(
        'utf-8'))


def get_computing_metadata():
    return json.loads(urlopen(
        'https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/metadata/v0.1/computing-metadata.json').read().decode(
        'utf-8'))


def get_computing_ddo():
    return _get_asset(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/compute/v0.1/ddo.computing.json")


def get_registered_ddo(nevermined_instance, account):
    metadata = get_metadata()
    metadata['main']['files'][0]['checksum'] = str(uuid.uuid4())
    ddo = nevermined_instance.assets.create(metadata, account)
    return ddo


def get_registered_with_psk(nevermined_instance, account, auth_method):
    metadata = get_ddo_sample()
    ddo = nevermined_instance.assets.create(metadata.metadata, account,
                                            authorization_type=auth_method)
    return ddo


def log_event(event_name):
    def _process_event(event):
        print(f'Received event {event_name}: {event}')

    return _process_event


def get_rsa_private_key_file():
    return os.getenv('RSA_PRIVKEY_FILE', '')


def get_rsa_public_key_file():
    return os.getenv('RSA_PUBKEY_FILE', '')


def get_metadata():
    return json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs"
        "/examples/metadata/v0.1/metadata1.json").read().decode(
        'utf-8'))


def setup_logging(default_path='logging.yaml', default_level=logging.INFO, env_key='LOG_CFG'):
    """Logging setup."""
    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, 'rt') as file:
            try:
                config = yaml.safe_load(file.read())
                logging.config.dictConfig(config)
                coloredlogs.install()
                logging.info(f'Logging configuration loaded from file: {path}')
            except Exception as ex:
                print(ex)
                print('Error in Logging Configuration. Using default configs')
                logging.basicConfig(level=default_level)
                coloredlogs.install(level=default_level)
    else:
        logging.basicConfig(level=default_level)
        coloredlogs.install(level=default_level)



import pytest
from common_utils_py.metadata.metadata import Metadata

from nevermined_sdk_py import ConfigProvider

metadata_provider = Metadata(ConfigProvider.get_config().metadata_url)


def test_get_service_endpoint():
    assert metadata_provider.get_service_endpoint() == f'{metadata_provider.url}/' + '{did}'


def test_publish_valid_ddo(asset1):
    metadata_provider.publish_asset_ddo(asset1)
    assert metadata_provider.get_asset_ddo(asset1.did)
    metadata_provider.retire_asset_ddo(asset1.did)


def test_publish_invalid_ddo():
    with pytest.raises(AttributeError):
        metadata_provider.publish_asset_ddo({})


def test_publish_ddo_already_registered(asset1):
    metadata_provider.publish_asset_ddo(asset1)
    with pytest.raises(ValueError):
        metadata_provider.publish_asset_ddo(asset1)
    metadata_provider.retire_asset_ddo(asset1.did)


def test_get_asset_ddo_for_not_registered_did():
    invalid_did = 'did:nv:not_valid'
    with pytest.raises(ValueError):
        metadata_provider.get_asset_ddo(invalid_did)


def test_get_asset_metadata(asset1):
    metadata_provider.publish_asset_ddo(asset1)
    metadata_dict = metadata_provider.get_asset_metadata(asset1.did)
    assert isinstance(metadata_dict, dict)
    assert 'main' in metadata_dict
    assert 'curation' in metadata_dict
    assert 'additionalInformation' in metadata_dict
    metadata_provider.retire_asset_ddo(asset1.did)


def test_get_asset_metadata_for_not_registered_did():
    invalid_did = 'did:nv:not_valid'
    with pytest.raises(ValueError):
        metadata_provider.get_asset_metadata(invalid_did)


def test_list_assets(asset1):
    num_assets = len(metadata_provider.list_assets())
    metadata_provider.publish_asset_ddo(asset1)
    assert len(metadata_provider.list_assets()) == (num_assets + 1)
    assert isinstance(metadata_provider.list_assets(), list)
    assert isinstance(metadata_provider.list_assets()[0], str)
    metadata_provider.retire_asset_ddo(asset1.did)


def test_list_assets_ddo(asset1):
    num_assets = len(metadata_provider.list_assets_ddo())
    metadata_provider.publish_asset_ddo(asset1)
    assert len(metadata_provider.list_assets_ddo()) == (num_assets + 1)
    assert isinstance(metadata_provider.list_assets_ddo(), dict)
    metadata_provider.retire_asset_ddo(asset1.did)


def test_update_ddo(asset1, asset2):
    metadata_provider.publish_asset_ddo(asset1)
    metadata_provider.update_asset_ddo(asset1.did, asset2)
    assert metadata_provider.get_asset_ddo(asset1.did).did == asset2.did
    assert metadata_provider.get_asset_ddo(asset1.did).metadata['main']['name'] != \
           asset1.metadata['main'][
               'name'], 'The name has not been updated correctly.'
    metadata_provider.retire_asset_ddo(asset1.did)


def test_update_with_not_valid_ddo(asset1):
    with pytest.raises(Exception):
        metadata_provider.update_asset_ddo(asset1.did, {})


def test_text_search(asset1, asset2):
    office_matches = 0
    metadata_provider.publish_asset_ddo(asset1)
    assert len(
        metadata_provider.text_search(text='Weather information', offset=10000)['results']) == (
                   office_matches + 1)

    text = 'UK'
    id_matches2 = len(metadata_provider.text_search(text=text, offset=10000)['results'])
    metadata_provider.publish_asset_ddo(asset2)
    assert len(metadata_provider.text_search(text=text, offset=10000)['results']) == (
            id_matches2 + 1)

    assert len(
        metadata_provider.text_search(text='Weather information', offset=10000)['results']) == (
                   office_matches + 2)
    metadata_provider.retire_asset_ddo(asset1.did)
    metadata_provider.retire_asset_ddo(asset2.did)


def test_text_search_invalid_query():
    with pytest.raises(Exception):
        metadata_provider.text_search(text='', offset='Invalid')


def test_query_search(asset1, asset2):
    num_matches = 0
    metadata_provider.publish_asset_ddo(asset1)

    assert len(metadata_provider.query_search(search_query={"query": {"type": ["dataset"]}},
                                              offset=10000)['results']) == (
                   num_matches + 1)

    metadata_provider.publish_asset_ddo(asset2)

    assert len(metadata_provider.query_search(search_query={"query": {"type": ["dataset"]}},
                                              offset=10000)['results']) == (
                   num_matches + 2)
    metadata_provider.retire_asset_ddo(asset1.did)
    metadata_provider.retire_asset_ddo(asset2.did)


def test_query_search_invalid_query():
    with pytest.raises(Exception):
        metadata_provider.query_search(search_query='')


def test_retire_ddo(asset1):
    n = len(metadata_provider.list_assets())
    metadata_provider.publish_asset_ddo(asset1)
    assert len(metadata_provider.list_assets()) == (n + 1)
    metadata_provider.retire_asset_ddo(asset1.did)
    assert len(metadata_provider.list_assets()) == n


def test_retire_all_ddos(asset1):
    metadata_provider.retire_all_assets()
    assert len(metadata_provider.list_assets()) == 0


def test_retire_not_published_did():
    with pytest.raises(Exception):
        metadata_provider.retire_asset_ddo('did:nv:not_registered')



from eth_utils import add_0x_prefix
from contracts_lib_py.web3_provider import Web3Provider
from common_utils_py.did import did_to_id


class Conditions:
    """Nevermined conditions class."""

    def __init__(self, keeper):
        self._keeper = keeper

    def lock_reward(self, agreement_id, amount, account):
        """
        Lock reward condition.

        :param agreement_id: id of the agreement, hex str
        :param amount: Amount of tokens, int
        :param account: Account
        :return: bool
        """
        self._keeper.token.token_approve(self._keeper.lock_reward_condition.address, amount,
                                         account)
        tx_hash = self._keeper.lock_reward_condition.fulfill(
            agreement_id, self._keeper.escrow_reward_condition.address, amount, account
        )
        receipt = self._keeper.lock_reward_condition.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)

    def grant_access(self, agreement_id, did, grantee_address, account):
        """
        Grant access condition.

        :param agreement_id: id of the agreement, hex str
        :param did: DID, str
        :param grantee_address: Address, hex str
        :param account: Account
        :return:
        """
        tx_hash = self._keeper.access_secret_store_condition.fulfill(
            agreement_id, add_0x_prefix(did_to_id(did)), grantee_address, account
        )
        receipt = self._keeper.access_secret_store_condition.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)

    def release_reward(self, agreement_id, amount, account):
        """
        Release reward condition.

        :param agreement_id: id of the agreement, hex str
        :param amount: Amount of tokens, int
        :param account: Account
        :return:
        """
        agreement_values = self._keeper.agreement_manager.get_agreement(agreement_id)
        consumer, provider = self._keeper.escrow_access_secretstore_template.get_agreement_data(
            agreement_id)
        owner = agreement_values.owner
        access_id, lock_id = agreement_values.condition_ids[:2]
        tx_hash = self._keeper.escrow_reward_condition.fulfill(
            agreement_id,
            amount,
            Web3Provider.get_web3().toChecksumAddress(owner),
            consumer,
            lock_id,
            access_id,
            account
        )
        receipt = self._keeper.escrow_reward_condition.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)

    def refund_reward(self, agreement_id, amount, account):
        """
        Refund reaward condition.

        :param agreement_id: id of the agreement, hex str
        :param amount: Amount of tokens, int
        :param account: Account
        :return:
        """
        return self.release_reward(agreement_id, amount, account)
